from dataclasses import asdict

from flask import abort, json, jsonify, request, send_from_directory, Blueprint

from paperead.env import env
from paperead.repository.materials import Material
from . import buildDocument


materials = Blueprint("materials", __name__)


@materials.route("/")
@materials.route("/index.json")
def allMaterial():
    return jsonify([id for id in env.repo])


@materials.route("/<id>/", methods=["GET"])
@materials.route("/<id>/index.json", methods=["GET"])
def getMaterial(id: str):
    if id not in env.repo:
        abort(404)
    item = env.repo[id]

    return jsonify(item)


@materials.route("/<id>/", methods=["DELETE"])
def deleteMaterial(id: str):
    if id not in env.repo:
        abort(404)
    del env.repo[id]
    return id


@materials.route("/", methods=["POST"])
def updateMaterial():
    obj = buildDocument(Material, request.get_json())

    env.repo.update(obj)
    return obj.id


@materials.route("/<id>/assets/<path:path>", methods=["GET"])
def assets(id: str, path: str):
    if id not in env.repo:
        abort(404)
    item = env.repo[id]
    return send_from_directory(item.assets, path)



from dataclasses import dataclass, field, asdict
import pathlib
from typing import Optional
from .repository.materials import MaterialRepository
import yaml

DEFAULT_SERVER_PORT = 3649
DEFAULT_BUILD_DIST = "./dist"
DEFAULT_BUILD_HOST = "empty"


@dataclass
class ServerConfig:
    port: int = DEFAULT_SERVER_PORT
    auth: str = ""
    readonly: bool = False


@dataclass
class BuildConfig:
    dist: str = DEFAULT_BUILD_DIST
    host: str = DEFAULT_BUILD_HOST


class Environment:
    def __init__(self, path: pathlib.Path) -> None:
        self.setPath(path)

    def setPath(self, path: pathlib.Path) -> None:
        self.path = path
        self.repo = MaterialRepository(path)

        self.configPath = self.path.joinpath(".paperead.yml")
        self.serverConfig = ServerConfig()
        self.buildConfig = BuildConfig()

        if self.configPath.exists() and self.configPath.is_file():
            text = self.configPath.read_text(encoding="utf-8")
            raw = yaml.safe_load(text)

            if "server" in raw:
                self.serverConfig = ServerConfig(**raw["server"])
            if "build" in raw:
                self.buildConfig = BuildConfig(**raw["build"])

    def generateConfigFile(self):
        self.configPath.write_text(yaml.safe_dump({
            "server": asdict(ServerConfig()),
            "build": asdict(BuildConfig())
        }))


env: Environment = Environment(pathlib.Path("."))

if __name__ == "__main__":
    env.setPath(pathlib.Path(
        __file__).parent.parent.parent.parent.joinpath("temp/materials"))
    print(env.serverConfig)



import pathlib

from flask import send_file, send_from_directory, Blueprint

wwwroot = pathlib.Path(__file__).parent.joinpath("wwwroot")

frontend = Blueprint("frontend", __name__)


@frontend.route("/", methods=["GET"])
@frontend.route("/<path:path>", methods=["GET"])
def wwwrootFiles(path: str = "index.html"):
    return send_from_directory(wwwroot, path)


@frontend.errorhandler(404)
def pageNotFound(error):
    return send_file(wwwroot.joinpath("index.html"))



from typing import Callable, Type

from paperead.repository import DocumentMetadata
from .. import app
from paperead import __version__

from flask import Blueprint, jsonify
from dateutil.parser import parse

api = Blueprint("api", __name__)


def buildDocument(type: Type, data: dict):
    metadata = data.pop("metadata")
    metadata["creation"] = parse(metadata["creation"])          # isoformat
    metadata["modification"] = parse(metadata["modification"])  # isoformat

    metaobj = DocumentMetadata(**metadata)
    return type(metaobj, **data)


@api.route("/")
@api.route("/index.json")
def apiinfo():
    return jsonify({
        "version": __version__
    })


def build():
    from .materials import materials
    from .notes import notes

    api.register_blueprint(materials, url_prefix="/materials")
    api.register_blueprint(notes, url_prefix="/materials")
    return api



from dataclasses import asdict

from dateutil.parser import parse
from flask import abort, json, jsonify, request, send_from_directory, Blueprint

from paperead.repository.notes import Note
from . import buildDocument

from paperead.env import env


notes = Blueprint("notes", __name__)


@notes.route("/<id>/notes/")
@notes.route("/<id>/notes/index.json")
def allNote(id: str):
    if id not in env.repo:
        abort(404)
    material = env.repo[id]
    return jsonify([item for item in material.notes])


@notes.route("/<id>/notes/<nid>/", methods=["GET"])
@notes.route("/<id>/notes/<nid>/index.json", methods=["GET"])
def getNote(id: str, nid: str):
    if id not in env.repo:
        abort(404)
    material = env.repo[id]
    if nid not in material.notes:
        abort(404)
    item = material.notes[nid]
    return jsonify(item)


@notes.route("/<id>/notes/<nid>/", methods=["DELETE"])
def deleteNote(id: str, nid: str):
    if id not in env.repo:
        abort(404)
    material = env.repo[id]
    if nid not in material.notes:
        abort(404)
    del material.notes[id]
    return id


@notes.route("/<id>/notes/", methods=["POST"])
def updateNote(id: str):
    if id not in env.repo:
        abort(404)

    obj = buildDocument(Note, request.get_json())

    material = env.repo[id]
    material.notes.update(obj)
    return obj.id



import datetime
import itertools
import os
import pathlib
import shutil
from dataclasses import asdict, dataclass, field
from typing import Dict, Iterator, List, Optional

import yaml
from dateutil.tz import tzlocal

from .. import fsutils
from . import Document, DocumentMetadata, DocumentRepository
from .notes import NoteRepository


@dataclass
class Material(Document):
    @property
    def notes(self) -> Optional[NoteRepository]:
        if hasattr(self, '_notes'):
            return self._notes
        else:
            return None

    @notes.setter
    def notes(self, value: NoteRepository):
        self._notes = value

    @property
    def assets(self) -> Optional[pathlib.Path]:
        if hasattr(self, '_assets'):
            return self._assets
        else:
            return None

    @assets.setter
    def assets(self, value: pathlib.Path):
        self._assets = value


class MaterialRepository(DocumentRepository[Material]):
    def __init__(self, root: pathlib.Path) -> None:
        super().__init__(root)

    @classmethod
    def __document__(cls, id: str, text: Optional[str] = None) -> Material:
        if text:
            return Material.fromText(id, text)
        else:
            return Material(DocumentMetadata(id), id)

    def __documentPath__(self, id: str) -> pathlib.Path:
        return self.root.joinpath(id.strip()).joinpath("description.md")

    def __documentGlob__(self) -> str:
        return "*/description.md"

    def __oniter__(self, path: pathlib.Path) -> str:
        return path.parent.stem

    def _getNotesPath(self, path: pathlib.Path) -> pathlib.Path:
        return path.parent.joinpath("notes")

    def _getAssetsPath(self, path: pathlib.Path) -> pathlib.Path:
        return path.parent.joinpath("assets")

    def __postget__(self, path: pathlib.Path, item: Material) -> None:
        item.notes = NoteRepository(self._getNotesPath(path))
        item.assets = self._getAssetsPath(path)

    def __postset__(self, path: pathlib.Path, item: Material) -> None:
        subroot = path.parent
        fsutils.ensureDirectory(subroot.joinpath("assets"))
        fsutils.ensureDirectory(subroot.joinpath("notes"))

    def __postdel__(self, id: str, path: pathlib.Path) -> None:
        shutil.rmtree(path.parent)



import datetime
import itertools
import os
import pathlib
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field
from typing import Any, Generic, Iterator, Optional, TypeVar, Dict, List

import yaml
from dateutil.tz import tzlocal

from .. import fsutils


@dataclass
class DocumentMetadata:
    name: str
    creation: datetime.datetime = field(
        default_factory=lambda: datetime.datetime.now(tzlocal()))
    modification: datetime.datetime = field(
        default_factory=lambda: datetime.datetime.now(tzlocal()))
    targets: Dict[str, str] = field(default_factory=dict)
    tags: List[str] = field(default_factory=list)
    extra: Dict[str, str] = field(default_factory=dict)
    schema: str = ""

    def toText(self) -> str:
        return yaml.safe_dump(asdict(self), sort_keys=False).strip()

    @classmethod
    def fromText(cls, text: str):
        data = yaml.safe_load(text)
        return cls(**data)


@dataclass
class Document(ABC):
    metadata: DocumentMetadata
    id: str = ""
    content: str = ""

    @classmethod
    def __metadata__(cls, name: str = "", text: Optional[str] = None) -> DocumentMetadata:
        if text:
            return DocumentMetadata.fromText(text)
        else:
            return DocumentMetadata(name)

    def toText(self) -> str:
        return "\n".join(["---", self.metadata.toText(), "---", self.content])

    @classmethod
    def fromText(cls, id: str, text: str):
        lines = text.splitlines()
        try:
            frontIndex = lines.index("---")
            if frontIndex != 0:
                raise ValueError
            frontIndex = lines.index("---", 1)

            front = "\n".join(lines[1:frontIndex])
            content = "\n".join(lines[frontIndex+1:])
            metadata = cls.__metadata__(text=front)
        except ValueError:
            content = text
            metadata = cls.__metadata__(name=id)
        return cls(id=id, metadata=metadata, content=content)


TD = TypeVar("TD", bound=Document)


class DocumentRepository(ABC, Generic[TD]):
    def __init__(self, root: pathlib.Path) -> None:
        self.root = root

    @classmethod
    @abstractmethod
    def __document__(cls, id: str, text: Optional[str] = None) -> TD:
        pass

    def __documentPath__(self, id: str) -> pathlib.Path:
        return self.root.joinpath(f"{id.strip()}.md")

    def __documentGlob__(self) -> str:
        return "*.md"

    def __postget__(self, path: pathlib.Path, item: TD) -> None:
        pass

    def __postset__(self, path: pathlib.Path, item: TD) -> None:
        pass

    def __postdel__(self, id: str, path: pathlib.Path) -> None:
        pass

    def __oniter__(self, path: pathlib.Path) -> str:
        return path.stem

    def __contains__(self, item: str) -> bool:
        path = self.__documentPath__(item)
        return path.exists() and path.is_file()

    def __getitem__(self, key: str) -> TD:
        if key in self:
            path = self.__documentPath__(key)
            result = self.__document__(
                key, path.read_text(encoding="utf-8"))
            self.__postget__(path, result)
            return result
        else:
            raise KeyError(key)

    def __setitem__(self, key: str, value: Optional[TD]) -> None:
        if value:
            value.id = key
            self.update(value)
        else:
            self.create(key)

    def __delitem__(self, key: str) -> None:
        if key in self:
            path = self.__documentPath__(key)
            os.remove(path)
            self.__postdel__(key, path)

    def __iter__(self) -> Iterator[str]:
        for item in self.root.glob(self.__documentGlob__()):
            yield self.__oniter__(item)

    def __len__(self) -> int:
        return sum((1 for _ in self.root.glob(self.__documentGlob__())))

    def create(self, id: str) -> None:
        if id in self:
            raise Exception(f"Document with id '{id}' exists.")
        result = self.__document__(id)
        self.update(result)

    def update(self, item: TD) -> None:
        path = self.__documentPath__(item.id)
        text = item.toText()
        fsutils.ensureFile(path, text)
        self.__postset__(path, item)



import pathlib

import click
import os
import shutil
import tornado.httpserver
import tornado.ioloop
import tornado.wsgi
from flask import request, abort


def buildApp():
    from . import app
    from .api import build as apiBuild
    from .frontend import frontend

    app.register_blueprint(apiBuild(), url_prefix="/api")
    app.register_blueprint(frontend, url_prefix="/")

    return app


def serve(debug: bool = False):
    from ..env import env
    
    app = buildApp()

    if env.serverConfig.auth:
        app.config["BASIC_AUTH_USERNAME"] = "admin"
        app.config["BASIC_AUTH_PASSWORD"] = str(env.serverConfig.auth)
        app.config["BASIC_AUTH_FORCE"] = True

        from flask_basicauth import BasicAuth
        BasicAuth(app)

    if env.serverConfig.readonly:
        @app.before_request
        def readonly():
            if request.method not in {"GET", "HEAD", "OPTIONS"}:
                abort(403)

    if debug:
        app.run(host="0.0.0.0", port=env.serverConfig.port, debug=debug)

    else:
        click.echo(f"Listening on port {env.serverConfig.port}...")
        click.echo(
            f"Visit http://localhost:{env.serverConfig.port} to Paperead.")

        container = tornado.wsgi.WSGIContainer(app)
        http_server = tornado.httpserver.HTTPServer(container)
        http_server.listen(env.serverConfig.port)
        tornado.ioloop.IOLoop.current().start()


def build():
    from ..env import env
    from .frontend import wwwroot

    repo = env.repo

    dist = pathlib.Path(env.buildConfig.dist)
    if not dist.is_absolute():
        dist = pathlib.Path(env.path).joinpath(dist)

    if dist.exists() and dist.is_dir():
        click.echo(f"Delete existed dist at {dist.absolute()}.")
        shutil.rmtree(dist)

    click.echo("Generating frontend.")

    shutil.copytree(wwwroot, dist)

    click.echo("Generating data.")

    app = buildApp()

    with app.test_client() as c:
        apidist = dist.joinpath("api")
        os.mkdir(apidist)

        apidist.joinpath("index.json").write_bytes(
            c.get(f"/api/").data)

        mdist = apidist.joinpath("materials")
        os.mkdir(mdist)

        mdist.joinpath("index.json").write_bytes(
            c.get(f"/api/materials/").data)

        for mid in repo:
            click.echo(f"  Generating material {mid}.")

            mdir = mdist.joinpath(mid)
            os.mkdir(mdir)

            mdir.joinpath("index.json").write_bytes(
                c.get(f"/api/materials/{mid}/").data)

            mdata = repo[mid]

            assets = mdata.assets
            if assets and assets.exists():
                click.echo(f"    Generating assets.")
                shutil.copytree(
                    assets,
                    mdir.joinpath("assets"))

            click.echo(f"    Generating notes.")

            ndist = mdir.joinpath("notes")
            os.mkdir(ndist)
            ndist.joinpath("index.json").write_bytes(
                c.get(f"/api/materials/{mid}/notes/").data)

            nrepo = mdata.notes

            for nid in nrepo:
                click.echo(f"      Generating note {nid}.")

                ndir = ndist.joinpath(nid)
                os.mkdir(ndir)

                ndir.joinpath("index.json").write_bytes(
                    c.get(f"/api/materials/{mid}/notes/{nid}/").data)

    staticPath = pathlib.Path(__file__).parent.joinpath("static")

    if env.buildConfig.host == 'python':
        click.echo("Generating files for python host.")
        click.echo(f"Use 'python serve.py' in {dist} to serve.")

        shutil.copyfile(staticPath.joinpath("__main__.py"),
                        dist.joinpath("serve.py"))
    elif env.buildConfig.host == 'netlify':
        click.echo("Generating files for netlify host.")
        shutil.copyfile(staticPath.joinpath("netlify").joinpath(
            "netlify.toml"), dist.joinpath("netlify.toml"))



import datetime
import itertools
import os
import pathlib
from dataclasses import asdict, dataclass, field
from typing import Dict, Iterator, Optional

import yaml
from dateutil.tz import tzlocal

from . import Document, DocumentMetadata, DocumentRepository


@dataclass
class Note(Document):
    pass


class NoteRepository(DocumentRepository[Note]):
    def __init__(self, root: pathlib.Path) -> None:
        super().__init__(root)

    @classmethod
    def __document__(cls, id: str, text: Optional[str] = None) -> Note:
        if text:
            return Note.fromText(id, text)
        else:
            return Note(DocumentMetadata(id), id)



# COPY from Python Standard Library: http.server

import copy
import datetime
import email.utils
import html
import http.client
import io
import mimetypes
import os
import posixpath
import select
import shutil
import socket # For gethostbyaddr()
import socketserver
import sys
import time
import urllib.parse
import contextlib
from functools import partial
from pathlib import Path

from http import HTTPStatus

import http.server
from http.server import test, ThreadingHTTPServer


class HttpRequestHandler(http.server.SimpleHTTPRequestHandler):
    extensions_map = {
        '': 'application/octet-stream',
        '.manifest': 'text/cache-manifest',
        '.html': 'text/html',
        '.png': 'image/png',
        '.jpg': 'image/jpg',
        '.svg':	'image/svg+xml',
        '.css':	'text/css',
        '.js':'application/x-javascript',
        '.wasm': 'application/wasm',
        '.json': 'application/json',
        '.xml': 'application/xml',
    }

    def do_GET(self):
        url_parts = urllib.parse.urlparse(self.path)
        request_file_path = Path(self.directory).joinpath(url_parts.path.strip("/"))

        ext = request_file_path.suffix

        if not request_file_path.is_file() and (ext == "" or ext not in self.extensions_map):
            # 404 handler
            self.path = 'index.html'

        return super().do_GET()


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--bind', '-b', metavar='ADDRESS',
                        help='Specify alternate bind address '
                             '[default: all interfaces]')
    parser.add_argument('--directory', '-d', default=os.getcwd(),
                        help='Specify alternative directory '
                        '[default:current directory]')
    parser.add_argument('port', action='store',
                        default=8000, type=int,
                        nargs='?',
                        help='Specify alternate port [default: 8000]')
    args = parser.parse_args()
    handler_class = partial(HttpRequestHandler,
                            directory=args.directory)

    # ensure dual-stack is not disabled; ref #38907
    class DualStackServer(ThreadingHTTPServer):
        def server_bind(self):
            # suppress exception when protocol is IPv4
            with contextlib.suppress(Exception):
                self.socket.setsockopt(
                    socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            return super().server_bind()

    test(
        HandlerClass=handler_class,
        ServerClass=DualStackServer,
        port=args.port,
        bind=args.bind,
    )


import os
import pathlib
from typing import Optional


def ensureDirectory(path: pathlib.Path) -> None:
    path = path.absolute()
    if path.exists() and path.is_dir():
        return

    os.makedirs(path, exist_ok=True)


def ensureFile(path: pathlib.Path, content: Optional[str] = None) -> None:
    path = path.absolute()
    if path.exists() and path.is_file():
        if content is not None:
            path.write_text(content)
        return

    ensureDirectory(path.parent)

    if content is None:
        content = ""

    path.write_text(content)



import logging
import pathlib
import time
from typing import Optional

import click
from click import BadArgumentUsage, BadOptionUsage, BadParameter
from click.exceptions import ClickException

from . import __version__
from .repository.materials import MaterialRepository

from .env import env, DEFAULT_BUILD_HOST, DEFAULT_BUILD_DIST, DEFAULT_SERVER_PORT


def setWorkingDirectory(wdir: pathlib.Path):
    env.setPath(wdir)


@click.command()
@click.option("-p", "--port", default=None, help="Port to serve.")
@click.option("-a", "--auth", default=None, help="Password to protect the website (Basic Auth with username 'admin'), empty for public access.")
@click.option("--readonly/--no-readonly", default=None, help="Readonly mode.")
@click.option("-d", "--debug", is_flag=True, default=False, help="Debug mode.")
def serve(port: Optional[int] = None, auth: Optional[str] = None, readonly: Optional[bool] = False, debug: bool = False) -> None:
    """Serve websites."""
    from .server import entrypoint
    if port:
        env.serverConfig.port = port
    if auth is not None:  # support empty string
        env.serverConfig.auth = auth
    if readonly is not None:
        env.serverConfig.readonly = readonly
    entrypoint.serve(debug)


@click.command()
@click.option("-h", "--host", default=None, type=click.Choice(["empty", "python", "netlify"]), help="Host for the static website.")
@click.option("-d", "--dist", default=None, help="Directory for building output.")
def build(host: Optional[str] = None, dist: Optional[str] = None) -> None:
    """Build static website."""
    from .server import entrypoint
    if host:
        env.buildConfig.host = host
    if dist:
        env.buildConfig.dist = dist
    entrypoint.build()


@click.command()
def init() -> None:
    """Initialize config files of paperead."""
    env.generateConfigFile()


@click.command()
@click.argument("id")
@click.option("-N", "--note", default="", help="Create a new note with specified id.")
def new(id: str, note: str = "") -> None:
    """Create a new material or note."""
    if note:
        if id not in env.repo:
            raise ClickException(f"Not found material with id '{id}'.")

        material = env.repo[id]
        if note in material.notes:
            raise ClickException(
                f"Note with id '{note}' exists for material with id '{id}'.")

        material.notes.create(note)

        click.echo(
            f"Created note with id '{note}' for material with id '{id}'.")
    else:
        if id in env.repo:
            raise ClickException(f"Material with id '{id}' exists.")

        env.repo.create(id)

        click.echo(f"Created material with id '{id}'.")


@click.command()
@click.argument("id")
@click.option("-N", "--note", default="", help="Remove a note with specified id.")
def rm(id: str, note: str = "") -> None:
    """Remove a material or note."""

    if note:
        if id not in env.repo:
            raise ClickException(f"Not found material with id '{id}'.")

        material = env.repo[id]
        if note not in material.notes:
            raise ClickException(
                f"Not found note with id '{note}' for material with id '{id}'.")

        del material.notes[note]

        click.echo(
            f"Removed note with id '{note}' for material with id '{id}'.")
    else:
        if id not in env.repo:
            raise ClickException(f"Not found material with id '{id}'.")

        del env.repo[id]

        click.echo(f"Removed material with id '{id}'.")


@click.command()
@click.argument("id", default="")
def list(id: str = "") -> None:
    """List materials and notes."""

    if id:
        if id not in env.repo:
            raise ClickException(f"Not found material with id '{id}'.")

        material = env.repo[id]

        for item in material.notes:
            click.echo(item)
    else:
        for item in env.repo:
            click.echo(item)


@click.group(invoke_without_command=True)
@click.pass_context
@click.option('-D', '--directory', type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path), default=".", help="Path to working directory.")
@click.option('-v', '--verbose', count=True, default=0, type=click.IntRange(0, 4))
@click.option("--version", is_flag=True, default=False, help="Show the version.")
def main(ctx=None, directory: pathlib.Path = ".", verbose: int = 0, version: bool = False) -> None:
    """Paperead (https://github.com/StardustDL/paperead)"""

    # click.echo(f"Welcome to Paperead v{__version__}!")

    setWorkingDirectory(directory)

    logger = logging.getLogger("Cli-Main")

    loggingLevel = {
        0: logging.ERROR,
        1: logging.WARNING,
        2: logging.INFO,
        3: logging.DEBUG,
        4: logging.NOTSET
    }[verbose]

    logging.basicConfig(level=loggingLevel)

    logger.debug(f"Logging level: {loggingLevel}")

    logger.info(f"Working directory: {click.format_filename(env.path)}")

    if version:
        click.echo(f"Paperead v{__version__}")
        exit(0)


main.add_command(serve)
main.add_command(build)
main.add_command(init)
main.add_command(new)
main.add_command(rm)
main.add_command(list)

if __name__ == '__main__':
    main()



import mimetypes
import pathlib
from typing import Optional

from flask import Flask

from flask_cors import CORS

mimetypes.add_type("text/css", ".css")
mimetypes.add_type("text/javascript", ".js")
mimetypes.add_type("text/javascript", ".mjs")

app = Flask(__name__)
CORS(app)



import pathlib

from paperead.repository.materials import MaterialRepository

if __name__ == "__main__":
    from . import entrypoint
    from ..env import env
    env.setPath(pathlib.Path(__file__).parent.parent.parent.parent.parent.joinpath(
        "temp").joinpath("materials"))
    entrypoint.serve(debug=True)
